import os
from typing import Any, TypeAlias

from boto3 import Session
from obstore.auth.boto3 import Boto3CredentialProvider
from obstore.store import (
    AzureStore,
    GCSStore,
    HTTPStore,
    LocalStore,
    MemoryStore,
    S3Store,
)
from sqlalchemy.engine.url import make_url
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.pool import StaticPool

ObjectStore: TypeAlias = (
    AzureStore | GCSStore | HTTPStore | S3Store | LocalStore | MemoryStore
)


class Config:
    def __init__(self):
        # db
        self.DATABASE_URL = os.getenv(
            "ICESTREAM_DATABASE_URL", "sqlite+aiosqlite:///:memory:"
        )
        self.async_session_factory: async_sessionmaker[AsyncSession] | None = None
        self.engine = None

        # obj store
        self.OBJECT_STORE_PROVIDER = os.getenv(
            "ICESTREAM_OBJECT_STORE_PROVIDER", "memory"
        )
        self.WAL_BUCKET = os.getenv("ICESTREAM_WAL_BUCKET", "icestream-wal")
        self.WAL_BUCKET_PREFIX = os.getenv("ICESTREAM_WAL_BUCKET_PREFIX", "")
        self.S3_ENDPOINT_URL = os.getenv("ICESTREAM_S3_ENDPOINT_URL")
        self.REGION = os.getenv("ICESTREAM_REGION", "us-east-1")
        self.MAX_IN_FLIGHT_FLUSHES = int(
            os.getenv("ICESTREAM_MAX_IN_FLIGHT_FLUSHES", "3")
        )
        self.store: ObjectStore = MemoryStore()

        # wal
        self.FLUSH_INTERVAL = int(os.getenv("ICESTREAM_FLUSH_INTERVAL", 2))
        self.FLUSH_SIZE = int(os.getenv("ICESTREAM_FLUSH_SIZE", 100 * 1024 * 1024))

        self.create_engine()
        self.create_store()

    def create_engine(self):
        url = make_url(self.DATABASE_URL)

        engine_options: dict[str, Any] = {
            "echo": True,
            "future": True,
        }

        if url.drivername.startswith("sqlite"):
            # Assign separately to keep type checkers happy
            engine_options["connect_args"] = {"check_same_thread": False}
            engine_options["poolclass"] = StaticPool

        elif url.drivername.startswith("postgresql"):
            if not url.drivername.startswith("postgresql+asyncpg"):
                url = url.set(drivername="postgresql+asyncpg")
        else:
            raise ValueError(f"Unsupported database dialect: {url.drivername}")

        self.engine = create_async_engine(url, **engine_options)
        self.async_session_factory = async_sessionmaker(
            bind=self.engine,
            expire_on_commit=False,
            class_=AsyncSession,
        )

    def create_store(self):
        bucket_path = self.WAL_BUCKET + self.WAL_BUCKET_PREFIX
        region = self._get_region()
        endpoint = self._get_endpoint()
        if self.OBJECT_STORE_PROVIDER == "aws":
            store_kwargs: dict[str, Any] = {"region": region}
            if endpoint is not None:
                store_kwargs["endpoint"] = endpoint
            # if any env var starts with AWS_, assume that we should get credentials that way
            if not any(key.startswith("AWS_") for key in os.environ):
                session = Session()
                credential_provider = Boto3CredentialProvider(session)
                store_kwargs["credential_provider"] = credential_provider
            self.store = S3Store(bucket_path, **store_kwargs)

    def _get_region(self):
        return os.getenv("AWS_DEFAULT_REGION") or os.getenv("AWS_REGION") or self.REGION

    def _get_endpoint(self):
        return (
            os.getenv("AWS_ENDPOINT")
            or os.getenv("AWS_ENDPOINT_URL")
            or self.S3_ENDPOINT_URL
        )



"""use partition topic_name and number in wal file offsets

Revision ID: d77fb56536a8
Revises: 5cc5497d09b5
Create Date: 2025-07-09 14:47:00.591448

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'd77fb56536a8'
down_revision: Union[str, None] = '5cc5497d09b5'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('wal_file_offsets', sa.Column('topic_name', sa.String(), nullable=False))
    op.add_column('wal_file_offsets', sa.Column('partition_number', sa.Integer(), nullable=False))
    op.create_index('ix_wal_file_offset_topic_partition', 'wal_file_offsets', ['topic_name', 'partition_number'], unique=False)
    op.drop_constraint(op.f('wal_file_offsets_partition_id_fkey'), 'wal_file_offsets', type_='foreignkey')
    op.create_foreign_key(None, 'wal_file_offsets', 'partitions', ['topic_name', 'partition_number'], ['topic_name', 'partition_number'], ondelete='CASCADE')
    op.drop_column('wal_file_offsets', 'partition_id')
    op.add_column('wal_files', sa.Column('uri', sa.Text(), nullable=False))
    op.add_column('wal_files', sa.Column('etag', sa.Text(), nullable=True))
    op.drop_column('wal_files', 'iceberg_commit_id')
    op.drop_column('wal_files', 's3_etag')
    op.drop_column('wal_files', 's3_uri')
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('wal_files', sa.Column('s3_uri', sa.TEXT(), autoincrement=False, nullable=False))
    op.add_column('wal_files', sa.Column('s3_etag', sa.TEXT(), autoincrement=False, nullable=True))
    op.add_column('wal_files', sa.Column('iceberg_commit_id', sa.TEXT(), autoincrement=False, nullable=True))
    op.drop_column('wal_files', 'etag')
    op.drop_column('wal_files', 'uri')
    op.add_column('wal_file_offsets', sa.Column('partition_id', sa.INTEGER(), autoincrement=False, nullable=False))
    op.drop_constraint(None, 'wal_file_offsets', type_='foreignkey')
    op.create_foreign_key(op.f('wal_file_offsets_partition_id_fkey'), 'wal_file_offsets', 'partitions', ['partition_id'], ['id'], ondelete='CASCADE')
    op.drop_index('ix_wal_file_offset_topic_partition', table_name='wal_file_offsets')
    op.drop_column('wal_file_offsets', 'partition_number')
    op.drop_column('wal_file_offsets', 'topic_name')
    # ### end Alembic commands ###



import io
import struct
from asyncio import StreamWriter
from dataclasses import dataclass
from typing import Awaitable, Callable

import structlog
from kio.index import load_request_schema, load_response_schema
from kio.schema.errors import ErrorCode
from kio.serial import entity_reader, entity_writer
from kio.static.constants import EntityType

from icestream.kafkaserver.handlers import KafkaHandler
from icestream.kafkaserver.handlers.api_versions import (
    ApiVersionsRequest,
    ApiVersionsRequestHeader,
    ApiVersionsResponse,
)
from icestream.kafkaserver.handlers.metadata import (
    MetadataRequest,
    MetadataRequestHeader,
    MetadataResponse,
)
from icestream.kafkaserver.handlers.produce import (
    ProduceRequest,
    ProduceRequestHeader,
    ProduceResponse,
)
from icestream.kafkaserver.messages import (
    CreatableTopicResult,
    CreateTopicsRequest,
    CreateTopicsRequestHeader,
    CreateTopicsResponse,
)

log = structlog.get_logger()

PRODUCE_API_KEY = 0
METADATA_API_KEY = 3
API_VERSIONS_API_KEY = 18
CREATE_TOPICS_API_KEY = 19


@dataclass
class RequestHandlerMeta:
    handler_func: Callable[
        [
            KafkaHandler,
            EntityType.header,
            EntityType.request,
            int,
            Callable[[EntityType.response], Awaitable[None]],
        ],
        Awaitable[None],
    ]
    error_response_func: Callable[
        [KafkaHandler, ErrorCode, str, EntityType.request, int], EntityType.response
    ]


api_compatibility: dict[int, tuple[int, int]] = {
    PRODUCE_API_KEY: (0, 8),
    METADATA_API_KEY: (0, 4),
    API_VERSIONS_API_KEY: (0, 4),
    CREATE_TOPICS_API_KEY: (0, 4),
}


async def handle_produce(
    handler: KafkaHandler,
    header: ProduceRequestHeader,
    req: ProduceRequest,
    api_version: int,
    respond: Callable[[ProduceResponse], Awaitable[None]],
) -> None:
    await handler.handle_produce_request(header, req, api_version, respond)


async def handle_metadata(
    handler: KafkaHandler,
    header: MetadataRequestHeader,
    req: MetadataRequest,
    api_version: int,
    respond: Callable[[MetadataResponse], Awaitable[None]],
) -> None:
    await handler.handle_metadata_request(header, req, api_version, respond)


async def handle_api_versions(
    handler: KafkaHandler,
    header: ApiVersionsRequestHeader,
    req: ApiVersionsRequest,
    api_version: int,
    respond: Callable[[ApiVersionsResponse], Awaitable[None]],
) -> None:
    await handler.handle_api_versions_request(header, req, api_version, respond)


async def handle_create_topics(
    handler: KafkaHandler,
    header: CreateTopicsRequestHeader,
    req: CreateTopicsRequest,
    api_version: int,
    respond: Callable[[CreateTopicsResponse], Awaitable[None]],
) -> None:
    await handler.handle_create_topics_request(header, req, api_version, respond)


def error_produce(
    handler: KafkaHandler,
    code: ErrorCode,
    msg: str,
    req: ProduceRequest,
    api_version: int,
) -> ProduceResponse:
    return handler.produce_request_error_response(code, msg, req, api_version)


def error_metadata(
    handler: KafkaHandler,
    code: ErrorCode,
    msg: str,
    req: MetadataRequest,
    api_version: int,
) -> MetadataResponse:
    return handler.metadata_request_error_response(code, msg, req, api_version)


def error_api_versions(
    handler: KafkaHandler,
    code: ErrorCode,
    msg: str,
    req: ApiVersionsRequest,
    api_version: int,
) -> ApiVersionsResponse:
    return handler.api_versions_request_error_response(code, msg, req, api_version)


def error_create_topics(
    handler: KafkaHandler,
    code: ErrorCode,
    msg: str,
    req: CreateTopicsRequest,
    api_version: int,
) -> CreateTopicsResponse:
    return handler.create_topics_request_error_response(code, msg, req, api_version)


request_map: dict[int, RequestHandlerMeta] = {
    PRODUCE_API_KEY: RequestHandlerMeta(
        handler_func=handle_produce,
        error_response_func=error_produce,
    ),
    METADATA_API_KEY: RequestHandlerMeta(
        handler_func=handle_metadata,
        error_response_func=error_metadata,
    ),
    API_VERSIONS_API_KEY: RequestHandlerMeta(
        handler_func=handle_api_versions,
        error_response_func=error_api_versions,
    ),
    CREATE_TOPICS_API_KEY: RequestHandlerMeta(
        handler_func=handle_create_topics,
        error_response_func=error_create_topics,
    ),
}


async def handle_kafka_request(
    api_key: int, buffer: bytes, handler: KafkaHandler, writer: StreamWriter
):
    if api_key not in request_map or api_key not in api_compatibility:
        return

    api_version = struct.unpack(">H", buffer[2:4])[0]
    buffer = io.BytesIO(buffer)

    log.info(f"got api key {api_key} with api version {api_version}")

    meta = request_map[api_key]
    min_vers, max_vers = api_compatibility[api_key]

    try:
        req_cls = load_request_schema(api_key, api_version)
        read_req = entity_reader(req_cls)
        read_req_header = entity_reader(req_cls.__header_schema__)

        req_header = read_req_header(buffer)
        req = read_req(buffer)

        resp_cls = load_response_schema(api_key, api_version)
        write_resp = entity_writer(resp_cls)
        write_resp_header = entity_writer(resp_cls.__header_schema__)

        response_header = resp_cls.__header_schema__(
            correlation_id=req_header.correlation_id
        )

        async def resp_func(resp: EntityType.response):
            resp_buffer = io.BytesIO()
            write_resp_header(resp_buffer, response_header)
            write_resp(resp_buffer, resp)
            resp_bytes = resp_buffer.getvalue()
            writer.write(len(resp_bytes).to_bytes(4, "big") + resp_bytes)
            await writer.drain()

        if not _is_in_supported_range(api_version, min_vers, max_vers):
            msg = f"supported versions for api key {api_key} are {min_vers} through {max_vers}"
            error_resp = meta.error_response_func(
                handler, ErrorCode.unsupported_version, msg, req, api_version
            )
            await resp_func(error_resp)
            return

        await meta.handler_func(handler, req_header, req, api_version, resp_func)

    except Exception as e:
        log.exception(
            "failed to handle kafka request",
            api_key=api_key,
            version=api_version,
            exception=e,
        )


def _is_in_supported_range(num: int, min_val: int, max_val: int) -> bool:
    return min_val <= num <= max_val



import struct
from dataclasses import dataclass
from io import BytesIO
from typing import Self


@dataclass
class KafkaRecordBatch:
    base_offset: int
    batch_length: int
    partition_leader_epoch: int
    magic: int
    crc: int
    attributes: int
    last_offset_delta: int
    base_timestamp: int
    max_timestamp: int
    producer_id: int
    producer_epoch: int
    base_sequence: int
    records_count: int
    records: bytes  # raw payload of [Record] section

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        buf = BytesIO(data)

        base_offset = struct.unpack(">q", buf.read(8))[0]
        batch_length = struct.unpack(">i", buf.read(4))[0]
        partition_leader_epoch = struct.unpack(">i", buf.read(4))[0]
        magic = struct.unpack(">b", buf.read(1))[0]
        crc = struct.unpack(">I", buf.read(4))[0]
        attributes = struct.unpack(">h", buf.read(2))[0]
        last_offset_delta = struct.unpack(">i", buf.read(4))[0]
        base_timestamp = struct.unpack(">q", buf.read(8))[0]
        max_timestamp = struct.unpack(">q", buf.read(8))[0]
        producer_id = struct.unpack(">q", buf.read(8))[0]
        producer_epoch = struct.unpack(">h", buf.read(2))[0]
        base_sequence = struct.unpack(">i", buf.read(4))[0]
        records_count = struct.unpack(">i", buf.read(4))[0]
        records = buf.read(
            batch_length - (buf.tell() - 12)
        )  # 12 = base_offset(8) + batch_length(4)

        return KafkaRecordBatch(
            base_offset,
            batch_length,
            partition_leader_epoch,
            magic,
            crc,
            attributes,
            last_offset_delta,
            base_timestamp,
            max_timestamp,
            producer_id,
            producer_epoch,
            base_sequence,
            records_count,
            records,
        )



from typing import Any, Awaitable, Callable, Protocol

from kio.schema.errors import ErrorCode

from icestream.kafkaserver.handlers.metadata import (
    MetadataRequest,
    MetadataRequestHeader,
    MetadataResponse,
)
from icestream.kafkaserver.handlers.produce import (
    ProduceRequest,
    ProduceRequestHeader,
    ProduceResponse,
)
from icestream.kafkaserver.messages import (
    ApiVersionsRequest,
    ApiVersionsRequestHeader,
    ApiVersionsResponse,
    CreateTopicsRequest,
    CreateTopicsRequestHeader,
    CreateTopicsResponse,
)


class KafkaHandler(Protocol):
    async def handle_produce_request(
        self,
        header: ProduceRequestHeader,
        req: ProduceRequest,
        api_version: int,
        callback: Callable[[ProduceResponse], Any],
    ):
        pass

    def produce_request_error_response(
        self,
        error_code: ErrorCode,
        error_message: str,
        req: ProduceRequest,
        api_version: int,
    ) -> ProduceResponse:
        pass

    async def handle_metadata_request(
        self,
        header: MetadataRequestHeader,
        req: MetadataRequest,
        api_version: int,
        callback: Callable[[MetadataResponse], Any],
    ):
        pass

    def metadata_request_error_response(
        self,
        error_code: ErrorCode,
        error_message: str,
        req: MetadataRequest,
        api_version: int,
    ) -> MetadataResponse:
        pass

    async def handle_api_versions_request(
        self,
        header: ApiVersionsRequestHeader,
        req: ApiVersionsRequest,
        api_version: int,
        callback: Callable[[ApiVersionsResponse], Awaitable[None]],
    ):
        pass

    def api_versions_request_error_response(
        self,
        error_code: ErrorCode,
        error_message: str,
        req: ApiVersionsRequest,
        api_version: int,
    ) -> ApiVersionsResponse:
        pass

    async def handle_create_topics_request(
        self,
        header: CreateTopicsRequestHeader,
        req: CreateTopicsRequest,
        api_version: int,
        callback: Callable[[CreateTopicsResponse], Awaitable[None]],
    ):
        pass

    def create_topics_request_error_response(
        self,
        error_code: ErrorCode,
        error_message: str,
        req: CreateTopicsRequest,
        api_version: int,
    ) -> CreateTopicsResponse:
        pass



import struct
import time
from io import BytesIO
from typing import List

from icestream.kafkaserver.protocol import KafkaRecordBatch
from icestream.kafkaserver.types import ProduceTopicPartitionData
from icestream.kafkaserver.wal import WALBatch, WALFile


def encode_varint(n: int) -> bytes:
    out = bytearray()
    while True:
        to_write = n & 0x7F
        n >>= 7
        if n:
            out.append(to_write | 0x80)
        else:
            out.append(to_write)
            break
    return bytes(out)


def encode_kafka_wal_file_with_offsets(
    batches: List[ProduceTopicPartitionData], broker_id: str
) -> tuple[bytes, list[dict]]:
    buf = BytesIO()
    offset_metadata = []

    # Write WAL file header
    buf.write(b"WAL1")  # Magic prefix
    buf.write(struct.pack(">B", 1))  # '>B': big-endian unsigned char (1 byte) - version
    buf.write(struct.pack(">Q", int(time.time() * 1000)))  # '>Q': big-endian unsigned long long (8 bytes) - flushed_at timestamp in ms

    # Write broker ID as length-prefixed UTF-8 string
    broker_bytes = broker_id.encode("utf-8")
    buf.write(encode_varint(len(broker_bytes)))  # Length of broker ID (varint)
    buf.write(broker_bytes)
    buf.write(encode_varint(len(batches)))  # Number of batches (varint)

    for batch in batches:
        topic_bytes = batch.topic.encode("utf-8")
        buf.write(encode_varint(len(topic_bytes)))  # Length of topic name (varint)
        buf.write(topic_bytes)
        buf.write(struct.pack(">i", batch.partition))  # '>i': big-endian signed int (4 bytes) - partition number

        record_batch_bytes_io = BytesIO()
        rb = batch.kafka_record_batch

        # Serialize Kafka record batch fields in order (see Kafka protocol spec)

        # '>q': big-endian signed long long (8 bytes) - base offset
        record_batch_bytes_io.write(struct.pack(">q", rb.base_offset))

        # '>i': big-endian signed int (4 bytes) - total batch length
        record_batch_bytes_io.write(struct.pack(">i", rb.batch_length))

        # '>i': big-endian signed int (4 bytes) - partition leader epoch
        record_batch_bytes_io.write(struct.pack(">i", rb.partition_leader_epoch))

        # '>b': big-endian signed char (1 byte) - magic byte (always 2)
        record_batch_bytes_io.write(struct.pack(">b", rb.magic))

        # '>I': big-endian unsigned int (4 bytes) - CRC32C checksum
        record_batch_bytes_io.write(struct.pack(">I", rb.crc))

        # '>h': big-endian signed short (2 bytes) - attributes bitfield
        record_batch_bytes_io.write(struct.pack(">h", rb.attributes))

        # '>i': big-endian signed int (4 bytes) - last offset delta
        record_batch_bytes_io.write(struct.pack(">i", rb.last_offset_delta))

        # '>q': big-endian signed long long (8 bytes) - base timestamp (ms)
        record_batch_bytes_io.write(struct.pack(">q", rb.base_timestamp))

        # '>q': big-endian signed long long (8 bytes) - max timestamp (ms)
        record_batch_bytes_io.write(struct.pack(">q", rb.max_timestamp))

        # '>q': big-endian signed long long (8 bytes) - producer ID
        record_batch_bytes_io.write(struct.pack(">q", rb.producer_id))

        # '>h': big-endian signed short (2 bytes) - producer epoch
        record_batch_bytes_io.write(struct.pack(">h", rb.producer_epoch))

        # '>i': big-endian signed int (4 bytes) - base sequence number
        record_batch_bytes_io.write(struct.pack(">i", rb.base_sequence))

        # '>i': big-endian signed int (4 bytes) - number of records
        record_batch_bytes_io.write(struct.pack(">i", rb.records_count))

        # Raw record data (variable length)
        record_batch_bytes_io.write(rb.records)

        record_batch_bytes = record_batch_bytes_io.getvalue()

        byte_start = buf.tell()
        buf.write(encode_varint(len(record_batch_bytes)))  # Length of batch (varint)
        buf.write(record_batch_bytes)
        byte_end = buf.tell()

        offset_metadata.append({
            "topic": batch.topic,
            "partition": batch.partition,
            "base_offset": rb.base_offset,
            "last_offset": rb.base_offset + rb.last_offset_delta,
            "byte_start": byte_start,
            "byte_end": byte_end,
        })

    return buf.getvalue(), offset_metadata

def decode_varint(buf: BytesIO) -> int:
    shift = 0
    result = 0
    while True:
        b = buf.read(1)
        if not b:
            raise EOFError("Unexpected EOF in varint")
        byte = b[0]
        result |= (byte & 0x7F) << shift
        if not (byte & 0x80):
            break
        shift += 7
    return result


def decode_kafka_wal_file(data: bytes) -> WALFile:
    buf = BytesIO(data)

    if buf.read(4) != b"WAL1":
        raise ValueError("Invalid WAL file magic")

    version = struct.unpack(">B", buf.read(1))[0]
    flushed_at = struct.unpack(">Q", buf.read(8))[0]
    broker_id_len = decode_varint(buf)
    broker_id = buf.read(broker_id_len).decode("utf-8")
    batch_count = decode_varint(buf)

    batches = []
    for _ in range(batch_count):
        topic_len = decode_varint(buf)
        topic = buf.read(topic_len).decode("utf-8")
        partition = struct.unpack(">i", buf.read(4))[0]
        batch_len = decode_varint(buf)
        record_batch_bytes = buf.read(batch_len)
        kafka_record_batch = KafkaRecordBatch.from_bytes(record_batch_bytes)
        batches.append(WALBatch(topic, partition, kafka_record_batch))

    return WALFile(version, flushed_at, broker_id, batches)



import asyncio
import datetime
import io
import struct
import uuid
from asyncio import Future, StreamReader, StreamWriter
from asyncio import Server as AsyncIOServer
from typing import Any, Callable, List, Sequence

import kio.schema.api_versions.v0 as api_v0
import kio.schema.api_versions.v1 as api_v1
import kio.schema.api_versions.v2 as api_v2
import kio.schema.api_versions.v3 as api_v3
import kio.schema.api_versions.v4 as api_v4
import kio.schema.metadata.v0 as metadata_v0
import kio.schema.metadata.v1 as metadata_v1
import kio.schema.metadata.v2 as metadata_v2
import kio.schema.metadata.v3 as metadata_v3
import kio.schema.metadata.v4 as metadata_v4
import kio.schema.metadata.v5 as metadata_v5
import kio.schema.metadata.v6 as metadata_v6
import kio.schema.produce.v0 as produce_v0
import kio.schema.produce.v1 as produce_v1
import kio.schema.produce.v2 as produce_v2
import kio.schema.produce.v3 as produce_v3
import kio.schema.produce.v4 as produce_v4
import kio.schema.produce.v5 as produce_v5
import kio.schema.produce.v6 as produce_v6
import kio.schema.produce.v7 as produce_v7
import kio.schema.produce.v8 as produce_v8
import structlog
from kio.index import load_payload_module
from kio.schema.errors import ErrorCode
from kio.schema.types import BrokerId, TopicName
from kio.serial.readers import read_int32
from kio.static.constants import EntityType
from kio.static.primitive import i16, i32, i32Timedelta, i64
from sqlalchemy import select, update
from sqlalchemy.orm import selectinload

from icestream.config import Config
from icestream.kafkaserver.handler import api_compatibility, handle_kafka_request
from icestream.kafkaserver.handlers import KafkaHandler
from icestream.kafkaserver.handlers.api_versions import (
    ApiVersionsRequest,
    ApiVersionsRequestHeader,
    ApiVersionsResponse,
)
from icestream.kafkaserver.handlers.metadata import (
    MetadataRequest,
    MetadataRequestHeader,
    MetadataResponse,
)
from icestream.kafkaserver.handlers.produce import (
    ProduceRequest,
    ProduceRequestHeader,
    ProduceResponse,
)
from icestream.kafkaserver.messages import (
    CreatableTopicResult,
    CreateTopicsRequest,
    CreateTopicsRequestHeader,
    CreateTopicsResponse,
)
from icestream.kafkaserver.protocol import KafkaRecordBatch
from icestream.kafkaserver.types import ProduceTopicPartitionData
from icestream.models import Partition, Topic

log = structlog.get_logger()


class Server:
    def __init__(self, config: Config, queue: asyncio.Queue[ProduceTopicPartitionData]):
        self.listener: AsyncIOServer | None = None
        self.config = config
        self.produce_queue = queue

    async def run(self, host: str = "127.0.0.1", port: int = 9092):
        try:
            self.listener = await asyncio.start_server(Connection(self), host, port)
            log.info(f"Server started listening on {host}:{port}")
            async with self.listener:
                await self.listener.serve_forever()
        except Exception as e:
            log.error(f"Error in server run: {e}")
            if self.listener:
                self.listener.close()


class Connection(KafkaHandler):
    def __init__(self, s: Server):
        self.server: Server = s

    async def __call__(self, reader: StreamReader, writer: StreamWriter) -> None:
        try:
            while True:
                msg_length_bytes = await reader.readexactly(4)
                msg_length = read_int32(io.BytesIO(msg_length_bytes))
                message = await reader.readexactly(msg_length)
                api_key = struct.unpack(">H", message[:2])[0]
                await handle_kafka_request(api_key, message, self, writer)

        except asyncio.IncompleteReadError as e:
            log.info(f"client disconnected with error: {e}")
        except Exception as e:
            log.error(f"error handling connection {e}")
        finally:
            try:
                writer.close()
                await writer.wait_closed()
                log.info("connection closed")
            except Exception as e:
                log.error(f"error closing writer: {e}")

    async def handle_produce_request(
        self,
        header: ProduceRequestHeader,
        req: ProduceRequest,
        api_version: int,
        callback: Callable[[ProduceResponse], Any],
    ):
        log.info(f"handling produce request {req}")
        topic_responses: List[produce_v8.response.TopicProduceResponse] = []

        for topic in req.topic_data:
            topic_name = topic.name
            partition_responses: List[produce_v8.response.PartitionProduceResponse] = []

            for partition in topic.partition_data:
                partition_idx = partition.index
                record_count = 0
                records = partition.records
                if records is not None:
                    record_count = int(struct.unpack(">I", records[57:61])[0])
                magic = records[16]
                error_code: ErrorCode | None = None
                if magic != 2:
                    error_code = ErrorCode.unsupported_for_message_format
                    partition_response = produce_v8.response.PartitionProduceResponse(
                        index=i32(partition_idx),
                        error_code=ErrorCode.none if error_code is None else error_code,
                        base_offset=i64(
                            -1
                        ),  # magic number wrong, so no offsets assigned
                        log_append_time=None,
                        log_start_offset=i64(-1),
                        record_errors=(),
                        error_message="wrong magic number",
                    )
                    partition_responses.append(partition_response)
                    continue

                log.info("produce", records=records[61:])

                log.info(
                    "produce",
                    topic=topic_name,
                    partition=partition_idx,
                    num_records=record_count,
                )

                if record_count == 0:
                    partition_response = produce_v8.response.PartitionProduceResponse(
                        index=i32(partition_idx),
                        error_code=ErrorCode.none if error_code is None else error_code,
                        base_offset=i64(-1),  # no records so no offsets assigned
                        log_append_time=None,
                        log_start_offset=i64(-1),
                        record_errors=(),
                        error_message=None,
                    )
                    partition_responses.append(partition_response)
                    continue

                # allocate the offsets
                async with self.server.config.async_session_factory() as session:
                    result = await session.execute(
                        update(Partition)
                        .where(
                            Partition.topic_name == topic_name,
                            Partition.partition_number == partition_idx,
                        )
                        .values(last_offset=Partition.last_offset + record_count)
                        .returning(
                            Partition.id,
                            Partition.last_offset,
                            Partition.log_start_offset,
                        )
                    )
                    await session.commit()
                row = result.first()
                # if the row is none, it's probably because there's no topic or partition for it
                if row is None:
                    error_code = ErrorCode.unknown_topic_or_partition
                    partition_response = produce_v8.response.PartitionProduceResponse(
                        index=i32(partition_idx),
                        error_code=ErrorCode.none if error_code is None else error_code,
                        base_offset=i64(
                            -1
                        ),  # likely no topic/partition, so no offsets assigned
                        log_append_time=None,
                        log_start_offset=i64(-1),
                        record_errors=(),
                        error_message="unknown topic or partition",
                    )
                    partition_responses.append(partition_response)
                    continue
                _, last_offset, log_start_offset = row
                first_offset = last_offset - record_count + 1

                # need to send the message batch over to the WALManager via self.server.produce_queue (asyncio.Queue)
                # but with a future so that this part that's creating the PartitionProduceResponse can wait for the response from the WALManager
                # but at the end need to append the PartitionProduceResponse to the `partition_responses` list.
                partition_flush_result_fut = Future()
                produce_topic_partition_data = ProduceTopicPartitionData(
                    topic=topic_name,
                    partition=partition_idx,
                    kafka_record_batch=KafkaRecordBatch.from_bytes(records),
                    flush_result=partition_flush_result_fut,
                )
                await self.server.produce_queue.put(produce_topic_partition_data)

                try:
                    await asyncio.wait_for(
                        partition_flush_result_fut,
                        timeout=self.server.config.FLUSH_INTERVAL * 2,
                    )
                    # success
                except asyncio.CancelledError:
                    error_code = ErrorCode.unknown_server_error
                except asyncio.TimeoutError:
                    error_code = ErrorCode.request_timed_out
                except Exception as e:
                    error_code = ErrorCode.unknown_server_error
                finally:
                    partition_response = produce_v8.response.PartitionProduceResponse(
                        index=i32(partition_idx),
                        error_code=ErrorCode.none if error_code is None else error_code,
                        base_offset=i64(first_offset if error_code is None else -1),
                        log_append_time=None,  # TODO
                        log_start_offset=i64(log_start_offset),
                        record_errors=(),
                        error_message=None,
                    )
                    partition_responses.append(partition_response)

            topic_response = produce_v8.response.TopicProduceResponse(
                name=topic_name,
                partition_responses=tuple(partition_responses),
            )
            topic_responses.append(topic_response)

        reference_response = produce_v8.response.ProduceResponse(
            responses=tuple(topic_responses),
            throttle_time=i32Timedelta.parse(datetime.timedelta(milliseconds=0)),
        )

        if api_version == 0:
            topics = []
            for topic in reference_response.responses:
                partitions = [
                    produce_v0.response.PartitionProduceResponse(
                        index=p.index,
                        error_code=p.error_code,
                        base_offset=p.base_offset,
                    )
                    for p in topic.partition_responses
                ]
                topics.append(
                    produce_v0.response.TopicProduceResponse(
                        name=topic.name,
                        partition_responses=tuple(partitions),
                    )
                )
            await callback(produce_v0.response.ProduceResponse(responses=tuple(topics)))

        elif api_version == 1:
            topics = []
            for topic in reference_response.responses:
                partitions = [
                    produce_v1.response.PartitionProduceResponse(
                        index=p.index,
                        error_code=p.error_code,
                        base_offset=p.base_offset,
                    )
                    for p in topic.partition_responses
                ]
                topics.append(
                    produce_v1.response.TopicProduceResponse(
                        name=topic.name,
                        partition_responses=tuple(partitions),
                    )
                )
            await callback(
                produce_v1.response.ProduceResponse(
                    responses=tuple(topics),
                    throttle_time=reference_response.throttle_time,
                )
            )

        elif api_version == 2:
            topics = []
            for topic in reference_response.responses:
                partitions = [
                    produce_v2.response.PartitionProduceResponse(
                        index=p.index,
                        error_code=p.error_code,
                        base_offset=p.base_offset,
                        log_append_time=p.log_append_time,
                    )
                    for p in topic.partition_responses
                ]
                topics.append(
                    produce_v2.response.TopicProduceResponse(
                        name=topic.name,
                        partition_responses=tuple(partitions),
                    )
                )
            await callback(
                produce_v2.response.ProduceResponse(
                    responses=tuple(topics),
                    throttle_time=reference_response.throttle_time,
                )
            )

        elif api_version == 3:
            topics = []
            for topic in reference_response.responses:
                partitions = [
                    produce_v3.response.PartitionProduceResponse(
                        index=p.index,
                        error_code=p.error_code,
                        base_offset=p.base_offset,
                        log_append_time=p.log_append_time,
                    )
                    for p in topic.partition_responses
                ]
                topics.append(
                    produce_v3.response.TopicProduceResponse(
                        name=topic.name,
                        partition_responses=tuple(partitions),
                    )
                )
            await callback(
                produce_v3.response.ProduceResponse(
                    responses=tuple(topics),
                    throttle_time=reference_response.throttle_time,
                )
            )

        elif api_version == 4:
            topics = []
            for topic in reference_response.responses:
                partitions = [
                    produce_v4.response.PartitionProduceResponse(
                        index=p.index,
                        error_code=p.error_code,
                        base_offset=p.base_offset,
                        log_append_time=p.log_append_time,
                    )
                    for p in topic.partition_responses
                ]
                topics.append(
                    produce_v4.response.TopicProduceResponse(
                        name=topic.name,
                        partition_responses=tuple(partitions),
                    )
                )
            await callback(
                produce_v4.response.ProduceResponse(
                    responses=tuple(topics),
                    throttle_time=reference_response.throttle_time,
                )
            )

        elif api_version == 5:
            topics = []
            for topic in reference_response.responses:
                partitions = [
                    produce_v5.response.PartitionProduceResponse(
                        index=p.index,
                        error_code=p.error_code,
                        base_offset=p.base_offset,
                        log_append_time=p.log_append_time,
                        log_start_offset=p.log_start_offset,
                    )
                    for p in topic.partition_responses
                ]
                topics.append(
                    produce_v5.response.TopicProduceResponse(
                        name=topic.name,
                        partition_responses=tuple(partitions),
                    )
                )
            await callback(
                produce_v5.response.ProduceResponse(
                    responses=tuple(topics),
                    throttle_time=reference_response.throttle_time,
                )
            )

        elif api_version == 6:
            topics = []
            for topic in reference_response.responses:
                partitions = [
                    produce_v6.response.PartitionProduceResponse(
                        index=p.index,
                        error_code=p.error_code,
                        base_offset=p.base_offset,
                        log_append_time=p.log_append_time,
                        log_start_offset=p.log_start_offset,
                    )
                    for p in topic.partition_responses
                ]
                topics.append(
                    produce_v6.response.TopicProduceResponse(
                        name=topic.name,
                        partition_responses=tuple(partitions),
                    )
                )
            await callback(
                produce_v6.response.ProduceResponse(
                    responses=tuple(topics),
                    throttle_time=reference_response.throttle_time,
                )
            )

        elif api_version == 7:
            topics = []
            for topic in reference_response.responses:
                partitions = [
                    produce_v7.response.PartitionProduceResponse(
                        index=p.index,
                        error_code=p.error_code,
                        base_offset=p.base_offset,
                        log_append_time=p.log_append_time,
                        log_start_offset=p.log_start_offset,
                    )
                    for p in topic.partition_responses
                ]
                topics.append(
                    produce_v7.response.TopicProduceResponse(
                        name=topic.name,
                        partition_responses=tuple(partitions),
                    )
                )
            await callback(
                produce_v7.response.ProduceResponse(
                    responses=tuple(topics),
                    throttle_time=reference_response.throttle_time,
                )
            )

        elif api_version == 8:
            await callback(reference_response)

        else:
            log.error(f"Unsupported produce version: {api_version}")

    def produce_request_error_response(
        self,
        error_code: ErrorCode,
        error_message: str,
        req: ProduceRequest,
        api_version: int,
    ) -> ProduceResponse:
        # v8 has error_message and record_errors in the PartitionProduceResponse, ignore it for now
        mod = load_payload_module(0, api_version, EntityType.response)
        responses = []
        for i, topic_data in enumerate(req.topic_data):
            partition_response = []
            for j, partition_data in enumerate(topic_data.partition_data):
                partition_produce_response = mod.PartitionProduceResponse(
                    index=partition_data.index,
                    error_code=error_code,
                    base_offset=i64(0),
                )
                partition_response.append(partition_produce_response)

            topic_produce_response = mod.TopicProduceResponse(
                name=topic_data.name, partition_responses=tuple(partition_response)
            )
            responses.append(topic_produce_response)
        resp = mod.ProduceResponse(responses=tuple(responses))
        return resp

    async def handle_metadata_request(
        self,
        header: MetadataRequestHeader,
        req: MetadataRequest,
        api_version: int,
        callback: Callable[[MetadataResponse], Any],
    ):
        log.info("handling metadata request", topics=[t.name for t in req.topics])

        async with self.server.config.async_session_factory() as session:
            topic_result: Sequence[Topic]
            if not req.topics:
                result = await session.execute(
                    select(Topic).options(selectinload(Topic.partitions))
                )
                topic_result = result.scalars().all()
            else:
                topic_names = [t.name for t in req.topics]
                result = await session.execute(
                    select(Topic)
                    .where(Topic.name.in_(topic_names))
                    .options(selectinload(Topic.partitions))
                )
                topic_result = result.scalars().all()

        # get brokers
        # since we're stateless we might be able to get away with spoofing a single broker
        # host would be the lb or k8s service or whatever
        # node id would always be 0
        # rack would always be None
        broker = metadata_v6.response.MetadataResponseBroker(
            node_id=i32(0), host="localhost", port=i32(9092), rack=None
        )

        # we need to respect the topic list passed in by the request
        # in our case it'll get passed to postgres, but an empty list means all of them

        topics: List[metadata_v6.response.MetadataResponseTopic] = []
        for topic in topic_result:
            partition_metadata = [
                metadata_v6.response.MetadataResponsePartition(
                    error_code=ErrorCode.none,
                    partition_index=i32(pidx.partition_number),
                    leader_id=i32(0),
                    replica_nodes=(i32(0),),
                    isr_nodes=(i32(0),),
                    offline_replicas=(),
                )
                for pidx in topic.partitions
            ]
            topics.append(
                metadata_v6.response.MetadataResponseTopic(
                    error_code=ErrorCode.none,
                    name=TopicName(topic.name),
                    is_internal=False,
                    partitions=tuple(partition_metadata),
                )
            )

        response = metadata_v6.response.MetadataResponse(
            throttle_time=i32Timedelta.parse(datetime.timedelta(milliseconds=0)),
            brokers=(broker,),
            cluster_id="test-cluster",
            controller_id=i32(0),
            topics=tuple(topics),
        )

        if api_version == 0:
            _broker = metadata_v0.response.MetadataResponseBroker(
                node_id=broker.node_id,
                host=broker.host,
                port=broker.port,
            )
            _topics = []
            for topic in topics:
                _partition_metadata = []
                for partition in topic.partitions:
                    _partition = metadata_v0.response.MetadataResponsePartition(
                        error_code=partition.error_code,
                        partition_index=partition.partition_index,
                        leader_id=partition.leader_id,
                        replica_nodes=partition.replica_nodes,
                        isr_nodes=partition.isr_nodes,
                    )
                    _partition_metadata.append(_partition)
                _topic = metadata_v0.response.MetadataResponseTopic(
                    error_code=topic.error_code,
                    name=topic.name,
                    partitions=tuple(_partition_metadata),
                )
                _topics.append(_topic)
            _response = metadata_v0.response.MetadataResponse(
                brokers=(_broker,), topics=tuple(_topics)
            )
            await callback(_response)

        elif api_version == 1:
            _broker = metadata_v1.response.MetadataResponseBroker(
                node_id=broker.node_id,
                host=broker.host,
                port=broker.port,
                rack=broker.rack,
            )
            _topics = []
            for topic in topics:
                _partition_metadata = []
                for partition in topic.partitions:
                    _partition = metadata_v1.response.MetadataResponsePartition(
                        error_code=partition.error_code,
                        partition_index=partition.partition_index,
                        leader_id=partition.leader_id,
                        replica_nodes=partition.replica_nodes,
                        isr_nodes=partition.isr_nodes,
                    )
                    _partition_metadata.append(_partition)
                _topic = metadata_v1.response.MetadataResponseTopic(
                    error_code=topic.error_code,
                    name=topic.name,
                    is_internal=topic.is_internal,
                    partitions=tuple(_partition_metadata),
                )
                _topics.append(_topic)
            _response = metadata_v1.response.MetadataResponse(
                brokers=(_broker,),
                topics=tuple(_topics),
                controller_id=BrokerId(_broker.node_id),
            )
            await callback(_response)

        elif api_version == 2:
            _broker = metadata_v2.response.MetadataResponseBroker(
                node_id=broker.node_id,
                host=broker.host,
                port=broker.port,
                rack=broker.rack,
            )
            _topics = []
            for topic in topics:
                _partition_metadata = []
                for partition in topic.partitions:
                    _partition = metadata_v2.response.MetadataResponsePartition(
                        error_code=partition.error_code,
                        partition_index=partition.partition_index,
                        leader_id=partition.leader_id,
                        replica_nodes=partition.replica_nodes,
                        isr_nodes=partition.isr_nodes,
                    )
                    _partition_metadata.append(_partition)
                _topic = metadata_v2.response.MetadataResponseTopic(
                    error_code=topic.error_code,
                    name=topic.name,
                    is_internal=topic.is_internal,
                    partitions=tuple(_partition_metadata),
                )
                _topics.append(_topic)
            _response = metadata_v2.response.MetadataResponse(
                brokers=(_broker,),
                topics=tuple(_topics),
                controller_id=BrokerId(_broker.node_id),
                cluster_id=response.cluster_id,
            )
            await callback(_response)

        elif api_version == 3:
            _broker = metadata_v3.response.MetadataResponseBroker(
                node_id=broker.node_id,
                host=broker.host,
                port=broker.port,
                rack=broker.rack,
            )
            _topics = []
            for topic in topics:
                _partition_metadata = []
                for partition in topic.partitions:
                    _partition = metadata_v3.response.MetadataResponsePartition(
                        error_code=partition.error_code,
                        partition_index=partition.partition_index,
                        leader_id=partition.leader_id,
                        replica_nodes=partition.replica_nodes,
                        isr_nodes=partition.isr_nodes,
                    )
                    _partition_metadata.append(_partition)
                _topic = metadata_v3.response.MetadataResponseTopic(
                    error_code=topic.error_code,
                    name=topic.name,
                    is_internal=topic.is_internal,
                    partitions=tuple(_partition_metadata),
                )
                _topics.append(_topic)
            _response = metadata_v3.response.MetadataResponse(
                brokers=(_broker,),
                topics=tuple(_topics),
                controller_id=BrokerId(_broker.node_id),
                cluster_id=response.cluster_id,
                throttle_time=response.throttle_time,
            )
            await callback(_response)

        elif api_version == 4:
            _broker = metadata_v4.response.MetadataResponseBroker(
                node_id=broker.node_id,
                host=broker.host,
                port=broker.port,
                rack=broker.rack,
            )
            _topics = []
            for topic in topics:
                _partition_metadata = []
                for partition in topic.partitions:
                    _partition = metadata_v4.response.MetadataResponsePartition(
                        error_code=partition.error_code,
                        partition_index=partition.partition_index,
                        leader_id=partition.leader_id,
                        replica_nodes=partition.replica_nodes,
                        isr_nodes=partition.isr_nodes,
                    )
                    _partition_metadata.append(_partition)
                _topic = metadata_v4.response.MetadataResponseTopic(
                    error_code=topic.error_code,
                    name=topic.name,
                    is_internal=topic.is_internal,
                    partitions=tuple(_partition_metadata),
                )
                _topics.append(_topic)
            _response = metadata_v4.response.MetadataResponse(
                brokers=(_broker,),
                topics=tuple(_topics),
                controller_id=BrokerId(_broker.node_id),
                cluster_id=response.cluster_id,
                throttle_time=response.throttle_time,
            )
            await callback(_response)

        elif api_version == 5:
            _broker = metadata_v5.response.MetadataResponseBroker(
                node_id=broker.node_id,
                host=broker.host,
                port=broker.port,
                rack=broker.rack,
            )
            _topics = []
            for topic in topics:
                _partition_metadata = []
                for partition in topic.partitions:
                    _partition = metadata_v5.response.MetadataResponsePartition(
                        error_code=partition.error_code,
                        partition_index=partition.partition_index,
                        leader_id=partition.leader_id,
                        replica_nodes=partition.replica_nodes,
                        isr_nodes=partition.isr_nodes,
                        offline_replicas=partition.offline_replicas,
                    )
                    _partition_metadata.append(_partition)
                _topic = metadata_v5.response.MetadataResponseTopic(
                    error_code=topic.error_code,
                    name=topic.name,
                    is_internal=topic.is_internal,
                    partitions=tuple(_partition_metadata),
                )
                _topics.append(_topic)
            _response = metadata_v5.response.MetadataResponse(
                brokers=(_broker,),
                topics=tuple(_topics),
                controller_id=BrokerId(_broker.node_id),
                cluster_id=response.cluster_id,
                throttle_time=response.throttle_time,
            )
            await callback(_response)

        elif api_version == 6:
            await callback(response)

        else:
            # unsupported - should be an error response
            pass

    def metadata_request_error_response(
        self,
        error_code: ErrorCode,
        error_message: str,
        req: MetadataRequest,
        api_version: int,
    ) -> MetadataResponse:
        # there's no point in returning brokers because there's no error code
        # similarly there's no point in populating the topic partitions
        # just populate the topics with the name and the error code
        # because the typing is weird, the req and api_version might not match
        mod = load_payload_module(3, api_version, EntityType.response)
        response_topic_class = mod.MetadataResponseTopic
        response_class = mod.MetadataResponse
        _topics = []
        for topic in req.topics:
            _topic = response_topic_class(name=topic.name, error_code=error_code)
            _topics.append(_topic)

        return response_class(brokers=(), topics=tuple(_topics))

    async def handle_api_versions_request(
        self,
        header: ApiVersionsRequestHeader,
        req: ApiVersionsRequest,
        api_version: int,
        callback: Callable[[ApiVersionsResponse], Any],
    ):
        versions = tuple(
            api_v4.response.ApiVersion(
                api_key=i16(api_key), min_version=i16(min_ver), max_version=i16(max_ver)
            )
            for api_key, (min_ver, max_ver) in api_compatibility.items()
        )

        response = api_v4.response.ApiVersionsResponse(
            error_code=ErrorCode.none,
            api_keys=versions,
            throttle_time=i32Timedelta.parse(datetime.timedelta(milliseconds=0)),
            supported_features=(),
            finalized_features_epoch=i64(-1),
            finalized_features=(),
            zk_migration_ready=False,
        )
        if api_version == 0:
            _versions = tuple(
                api_v0.response.ApiVersion(
                    api_key=_version.api_key,
                    min_version=_version.min_version,
                    max_version=_version.max_version,
                )
                for _version in response.api_keys
            )
            _response = api_v0.response.ApiVersionsResponse(
                error_code=response.error_code, api_keys=_versions
            )
            await callback(_response)
        elif api_version == 1:
            _versions = tuple(
                api_v1.response.ApiVersion(
                    api_key=_version.api_key,
                    min_version=_version.min_version,
                    max_version=_version.max_version,
                )
                for _version in response.api_keys
            )
            _response = api_v1.response.ApiVersionsResponse(
                error_code=response.error_code,
                api_keys=_versions,
                throttle_time=response.throttle_time,
            )
            await callback(_response)
        elif api_version == 2:
            _versions = tuple(
                api_v2.response.ApiVersion(
                    api_key=_version.api_key,
                    min_version=_version.min_version,
                    max_version=_version.max_version,
                )
                for _version in response.api_keys
            )
            _response = api_v2.response.ApiVersionsResponse(
                error_code=response.error_code,
                api_keys=_versions,
                throttle_time=response.throttle_time,
            )
            await callback(_response)
        elif api_version == 3:
            _versions = tuple(
                api_v3.response.ApiVersion(
                    api_key=_version.api_key,
                    min_version=_version.min_version,
                    max_version=_version.max_version,
                )
                for _version in response.api_keys
            )
            _response = api_v3.response.ApiVersionsResponse(
                error_code=response.error_code,
                api_keys=_versions,
                throttle_time=response.throttle_time,
                supported_features=response.supported_features,
                finalized_features_epoch=response.finalized_features_epoch,
                finalized_features=response.finalized_features,
                zk_migration_ready=response.zk_migration_ready,
            )
            await callback(_response)
        elif api_version == 4:
            await callback(response)
        else:
            log.error(f"unsupported api versions version: {api_version}")

    def api_versions_request_error_response(
        self,
        error_code: ErrorCode,
        error_message: str,
        req: ApiVersionsRequest,
        api_version: int,
    ) -> ApiVersionsResponse:
        mod = load_payload_module(18, api_version, EntityType.response)
        return mod.ApiVersionsResponse(
            error_code=error_code,
            api_keys=(),
        )

    async def handle_create_topics_request(
        self,
        header: CreateTopicsRequestHeader,
        req: CreateTopicsRequest,
        api_version: int,
        callback: Callable[[CreateTopicsResponse], Any],
    ):
        results = []

        for topic in req.topics:
            log.info(
                "create_topic", topic=topic.name, num_partitions=topic.num_partitions
            )

            result = CreatableTopicResult(
                name=topic.name,
                topic_id=uuid.uuid4(),
                error_code=ErrorCode.none,
                error_message=None,
                topic_config_error_code=i16(0),
                num_partitions=i32(topic.num_partitions),
                replication_factor=i16(topic.replication_factor),
                configs=None,  # configs not returned
            )
            results.append(result)

        response = CreateTopicsResponse(
            throttle_time=i32Timedelta.parse(datetime.timedelta(milliseconds=0)),
            topics=tuple(results),
        )
        await callback(response)

    def create_topics_request_error_response(
        self,
        error_code: ErrorCode,
        error_message: str,
        req: CreateTopicsRequest,
        api_version: int,
    ) -> CreateTopicsResponse:
        results = []

        for topic in req.topics:
            result = CreatableTopicResult(
                name=topic.name,
                topic_id=None,
                error_code=error_code,
                error_message=error_message,
                topic_config_error_code=i16(0),
                num_partitions=i32(-1),
                replication_factor=i16(-1),
                configs=None,
            )
            results.append(result)

        return CreateTopicsResponse(
            throttle_time=i32Timedelta.parse(datetime.timedelta(milliseconds=0)),
            topics=tuple(results),
        )



import asyncio
import time
import uuid
from io import BytesIO
from typing import List, Callable

from icestream.config import Config
from icestream.kafkaserver.types import ProduceTopicPartitionData
from icestream.kafkaserver.wal.serde import encode_kafka_wal_file_with_offsets
from icestream.logger import log
from icestream.models import WALFile, WALFileOffset


class WALManager:
    def __init__(self, config: Config, queue: asyncio.Queue[ProduceTopicPartitionData], time_source: Callable[[], float] = time.monotonic):
        self.config = config
        self.queue = queue
        self.time_source = time_source
        self.buffer: List[ProduceTopicPartitionData] = []
        self.buffer_size = 0
        self.last_flush_time = self.time_source()
        self.flush_semaphore = asyncio.Semaphore(self.config.MAX_IN_FLIGHT_FLUSHES)
        self.pending_flushes: set[asyncio.Task] = set()
        log.info(
            f"WALManager initialized: flush size={self.config.FLUSH_SIZE}, "
            f"interval={self.config.FLUSH_INTERVAL}, "
            f"max in-flight flushes={self.config.MAX_IN_FLIGHT_FLUSHES}"
        )

    async def run(self):
        log.info("WALManager started")
        try:
            while True:
                await self.run_once()
        except asyncio.CancelledError:
            log.info("WALManager run loop cancelled, flushing remaining buffer")

            for task in self.pending_flushes:
                if not task.done():
                    task.cancel()
            if self.pending_flushes:
                await asyncio.gather(*self.pending_flushes, return_exceptions=True)

            if self.buffer:
                await self._flush(self.buffer)
            raise
        finally:
            log.info("WALManager stopped")

    async def run_once(self, now: float | None = None):
        now = now or self.time_source()
        timeout = self.config.FLUSH_INTERVAL - (now - self.last_flush_time)
        try:
            item = await asyncio.wait_for(
                self.queue.get(), timeout=max(0, timeout)
            )
            self.buffer.append(item)
            self.buffer_size += item.kafka_record_batch.batch_length  # TODO
            if self.buffer_size >= self.config.FLUSH_SIZE:
                await self._launch_flush()
        except asyncio.TimeoutError:
            await self._launch_flush()

    async def _launch_flush(self):
        self.last_flush_time = time.monotonic()
        if not self.buffer:
            return
        batch_to_flush = self.buffer
        self.buffer = []
        self.buffer_size = 0

        flush_task = asyncio.create_task(self._flush(batch_to_flush))
        self.pending_flushes.add(flush_task)
        flush_task.add_done_callback(lambda t: self.pending_flushes.discard(t))


    async def _flush(self, batch_to_flush: List[ProduceTopicPartitionData]):
        try:
            async with self.flush_semaphore:
                encoded, offsets = encode_kafka_wal_file_with_offsets(
                    batch_to_flush, broker_id="foo"  # TODO
                )
                object_key = self._generate_object_key()
                log.info(
                    f"WALManager encoded {len(encoded)} bytes for {len(batch_to_flush)} batches"
                )

                put_result = await self.config.store.put_async(
                    path=object_key,
                    file=BytesIO(encoded),
                    use_multipart=len(encoded) > 5 * 1024 * 1024,
                )
                log.info(f"put_result {put_result}")

                async with self.config.async_session_factory() as session:
                    total_records = sum(o["last_offset"] - o["base_offset"] + 1 for o in offsets)

                    wal_file = WALFile(
                        uri=f"{self.config.WAL_BUCKET}{"/" + self.config.WAL_BUCKET_PREFIX if self.config.WAL_BUCKET_PREFIX else ""}/{object_key}",
                        etag=getattr(put_result, "etag", None),
                        total_bytes=len(encoded),
                        total_messages=total_records,
                    )
                    session.add(wal_file)
                    await session.flush()  # get wal_file.id

                    for o in offsets:
                        wal_file_offset = WALFileOffset(
                            wal_file_id=wal_file.id,
                            topic_name=o["topic"],
                            partition_number=o["partition"],
                            base_offset=o["base_offset"],
                            last_offset=o["last_offset"],
                            byte_start=o["byte_start"],
                            byte_end=o["byte_end"],
                        )
                        session.add(wal_file_offset)

                    await session.commit()

                for item in batch_to_flush:
                    if not item.flush_result.done():
                        item.flush_result.set_result(True)

                log.info("WALManager flushed successfully")

        except asyncio.CancelledError:
            log.info("WALManager flush cancelled, setting futures to cancelled")
            for item in batch_to_flush:
                if not item.flush_result.done():
                    item.flush_result.cancel()

        except Exception as e:
            log.exception("WALManager flush failed")
            for item in batch_to_flush:
                if not item.flush_result.done():
                    item.flush_result.set_exception(e)

    @staticmethod
    def _generate_object_key() -> str:
        now = time.gmtime()
        ts = int(time.time() * 1000)
        uid = uuid.uuid4().hex
        bucket = uid[:2]
        return (
            f"wal/{now.tm_year:04}/{now.tm_mon:02}/{now.tm_mday:02}/"
            f"{now.tm_hour:02}/{now.tm_min:02}/{bucket}/{ts}-{uid}.wal"
        )



from kio.schema.api_versions.v4.request import (
    ApiVersionsRequest,
)
from kio.schema.api_versions.v4.request import (
    RequestHeader as ApiVersionsRequestHeader,
)
from kio.schema.api_versions.v4.response import (
    ApiVersion,
    ApiVersionsResponse,
)
from kio.schema.api_versions.v4.response import (
    ResponseHeader as ApiVersionsResponseHeader,
)
from kio.schema.create_topics.v7.request import (
    CreateTopicsRequest,
)
from kio.schema.create_topics.v7.request import (
    RequestHeader as CreateTopicsRequestHeader,
)
from kio.schema.create_topics.v7.response import (
    CreatableTopicResult,
    CreateTopicsResponse,
)
from kio.schema.create_topics.v7.response import (
    ResponseHeader as CreateTopicsResponseHeader,
)
from kio.schema.metadata.v12.request import (
    MetadataRequest,
    MetadataRequestTopic,
    TopicName,
)
from kio.schema.metadata.v12.request import (
    RequestHeader as MetadataRequestHeader,
)
from kio.schema.metadata.v12.response import (
    MetadataResponse,
    MetadataResponseBroker,
    MetadataResponsePartition,
    MetadataResponseTopic,
)
from kio.schema.metadata.v12.response import (
    ResponseHeader as MetadataResponseHeader,
)
from kio.schema.produce.v9.request import (
    PartitionProduceData,
    ProduceRequest,
    TopicProduceData,
)
from kio.schema.produce.v9.request import (
    RequestHeader as ProduceRequestHeader,
)
from kio.schema.produce.v9.response import (
    PartitionProduceResponse,
    ProduceResponse,
    TopicProduceResponse,
)
from kio.schema.produce.v9.response import (
    ResponseHeader as ProduceResponseHeader,
)
